/****************************************************************
 *           EXAMPLE  FOR  USING                                *       
 *           KASTL TRANSFORM ALGORITHM                          *
 *                                                              *
 ****************************************************************/

#include <iostream>
#include <numeric>
#include <vector>
#include <cmath>
#include <climits>
#include <limits>
#include <algorithm>

#include<athapascan-1>

typedef  int  val_t ;

/*To get the time*/
inline double gettime()
{
   return Util::WallTimer::gettime();
}


/*To print the execution of the program*/
void print_time(double t)
{
  static double shift = 0;
  double delay = gettime()-t;
  std::cout << a1::System::getRank() << ":-----------------------------------------" << std::endl;
  std::cout << a1::System::getRank() << ": time = " << delay-shift << " s" << std::endl;
  std::cout << a1::System::getRank() << ":-----------------------------------------" << std::endl;
  shift = delay;
}

/*Function to verify if the table has been well sorted*/
template<class T>
bool is_sorted(T output, int sz)
{
     bool is_sorted = true;
     for(int i=1; i < sz; i++) {
         if(output[i-1] > output[i]) {
             std::cout << "output[" << i-1 << "]=" << output[i-1] << " is not less than  ";
             std::cout << "output[" << i << "]=" << output[i] << ", " << std::endl;
             std::cout << "Then the table is not sorted well" << std::endl;
             is_sorted = false;
           break;
        }
     }
  return is_sorted;
}


/*The main structure where the kastl transform is call*/
struct doit {
   
   void operator()(int argc, char ** argv) {
      	    
      size_t size = (argc > 1)?atoi(argv[1]):100000000;
      int v = (argc > 2)?atoi(argv[2]):0;
       
      std::vector<val_t> in_data(size);
      
      std::cout << "size =" << size << std::endl;

       //initialization of input data   
       srand48(0);
       for(size_t i =0; i < size; i++) in_data[i] = lrand48(); //drand48() ; // initial value : put what you want
       
       std::cout << "***********Start of computation**********" << std::endl;
       std::cout << "The computation size is: " << size << std::endl;   
       
       //start of the computation
       double time = gettime();  
       // The stl sort
        std::sort(in_data.begin(), in_data.end()); 

       //print the execution time*/
       print_time(time);

       if(is_sorted(in_data, size))
         std::cout << ":The table has been well sorted.........." << std::endl;

 }
};

/*The main function*/
int main(int argc,char**argv){

    a1::Community com = a1::System::join_community(argc,argv);
    a1::ForkMain<doit>() (argc,argv);
    com.leave();
    a1::System::terminate();

}

