/*Author : D Traore
 *
*/
//#ifndef _CILK_ADAPT_H_
//#define _CILK_ADAPT_H_
/*
 * this files groups definition common to all adaptive algorithm
 * Author: D. Traore, S. Guelton
 */

#include <algorithm>

#include <iostream>
#include <assert.h>
#include <vector>
#include <list>

#include <functional>

#include <cilk.h>

namespace adapt {

inline uint log_2 ( uint a )  // returns [__log_2 a]+1
{
        uint res = 0 ;
        for ( ; a != 0 ; a >>=1, res++ ) ;
        return res ;
}

/*
 * forward declaration
 */
template<class T> class smart_ptr;
class WorkAdapt;
class FinalizeWork;
typedef smart_ptr<WorkAdapt*> WorkAdaptPtr;
typedef smart_ptr<FinalizeWork*> FinalizeWorkPtr;

/*
 * encapsulate a pointer and handles communication and memory allocation
 */

struct unmanaged {};

template <class T>
class smart_ptr<T*>
{
        T* _ptr;
        //Util::Atomic *_ref;

        void add_ref() { 
          //_ref->incr();
        }

        void rm_ref(){
            /*if( _ref->decr_and_return() == 0)
            {
                delete _ptr;
                delete _ref;
            }*/
        }

        friend class WorkAdapt;
        friend class FinalizeWork;

    public:
        smart_ptr():
            _ptr(0) /*, _ref(0)*/
        {}

        smart_ptr(T* ptr):
            _ptr(ptr)/*, _ref(new Util::Atomic())*/
        {
            //_ref->initialize();
            //add_ref();
        }
        smart_ptr(T* ptr, unmanaged ):
            _ptr(ptr) /*, _ref(0)*/
        {
        }

        smart_ptr(const smart_ptr<T*>& sp)
            :_ptr(sp._ptr) /*, _ref(sp._ref)*/
        {
            //if(_ref) add_ref();
        }

        smart_ptr<T*>& operator=(const smart_ptr<T*>& sp)
        {
            if( &sp != this)
            {
                //if(_ref) rm_ref();
                _ptr=sp._ptr;
                //_ref=sp._ref;
                //if(_ref) add_ref();
            }
            return *this;
        }

        ~smart_ptr()
        {
            //if( _ref) rm_ref();
        }

        T* operator->() { return _ptr; }
        const T* operator->() const { return _ptr; }
        T& operator*() { return *_ptr; }
        const T& operator*() const { return *_ptr; }
};


class NextJump {
  
   public :
     NextJump() { }
     virtual void init_work() const { }
     virtual void init_work() {}
};

class FinalizeWork {

    pthread_mutex_t _mutex;

    protected:

        virtual bool extract_nextseq() = 0 ;
        virtual bool extract_par( FinalizeWork*& ) = 0 ;
        virtual void localcompute() =0 ;

   public :

     bool _finalize_signal;

     FinalizeWork()
     {
         pthread_mutex_init(&_mutex, NULL);
         _finalize_signal = false; 
     }
     
     bool ExtractNextSeq()
     {
           pthread_mutex_lock(&_mutex);
             bool local_extract_is_possible = extract_nextseq();
           pthread_mutex_unlock(&_mutex);
         return local_extract_is_possible;
     }

     bool ExtractPar( FinalizeWorkPtr& wp)

     {
            pthread_mutex_lock(&_mutex);
               bool steal_extract_is_possible = extract_par(wp._ptr);
            pthread_mutex_unlock(&_mutex);
          return steal_extract_is_possible;
     }

     void LocalCompute() 
     {
       localcompute();
     }

   friend class WorkAdapt;

};


/*
 * main interface for adaptive works 
 */
class WorkAdapt
{
        pthread_mutex_t  _mutex ;
        std::list< WorkAdapt* > thief_list;
        int not_finpar;
        bool _signalMerge;
        bool _signalJump;
        bool _progress;
        bool _update_work;
        WorkAdapt* _thieft_work;
        FinalizeWork* _fw; 

    protected:

        virtual bool extract_nextseq() = 0 ;
        virtual bool extract_par( WorkAdapt/*Ptr*/*& ) = 0 ;
        virtual void localcompute() =0 ;
        virtual void join(const WorkAdapt* ) { };
        virtual bool extract_next_macrostep() { return false; };
        virtual void setjump(NextJump*& ) = 0;
        virtual void jump(const NextJump* )= 0;
        virtual void get_main_result(const WorkAdapt* ) {  };  
        virtual void get_thieft_result() {  };
        virtual void get_finalize_work(FinalizeWork*& ) {  };
        virtual inline bool is_finalize_work() const {  return false; }
        virtual bool is_empty() = 0 ;
 
    public:
        bool _finalize_signal;

        
        virtual inline bool  MustJump() const { return false; }
        virtual inline bool  MainJump() const { return true; } 

        WorkAdapt() 
        { 
           pthread_mutex_init(&_mutex, NULL); 
            not_finpar = 1;
           _signalMerge = false; 
           _signalJump = false; 
            _progress = true;
           _update_work = false;
            _finalize_signal = false;

        }

        virtual ~WorkAdapt() {}

        bool ExtractNextSeq()
        {
           pthread_mutex_lock(&_mutex);
            bool local_extract_is_possible = extract_nextseq();
            if(!local_extract_is_possible) not_finpar = 0;
           pthread_mutex_unlock(&_mutex);
          return local_extract_is_possible; 
        }

        bool ExtractPar( WorkAdaptPtr& wp)
        {
             pthread_mutex_lock(&_mutex);
               bool steal_extract_is_possible = extract_par(wp._ptr);
               if(steal_extract_is_possible) thief_list.push_front(wp._ptr);
             pthread_mutex_unlock(&_mutex);
           return steal_extract_is_possible;
        }

       
        void LocalCompute()
        {
            localcompute();
        }

        void GetThieftWork( WorkAdapt*& thieftwork) {
           WorkAdapt *wa = thief_list.front();
           thieftwork = wa;
           thief_list.pop_front();
        }


        void Join(WorkAdapt*& jw )
        {
             join(jw);
        }
          
        void merge() {
             Join(_thieft_work);
        }


       bool getResultToPreempt(NextJump*& j)
       {
         bool go_to_next_thief = (nb_thief()!=0);
         pthread_mutex_lock(&_mutex);
          if(_signalMerge) {
            go_to_next_thief = false;
            _signalJump = true;
             not_finpar = 0;
          } else  not_finpar = 1;
         pthread_mutex_unlock(&_mutex);
         if(go_to_next_thief) { return mergeWithNextThiefResult(j) ; }
         else  {  not_finpar = 0; return false; }
      }

     bool mergeWithNextThiefResult(NextJump*& j)
      {   
         bool tmp;
         pthread_mutex_lock(&_mutex);
            WorkAdapt *wa = thief_list.front();
            _thieft_work = wa;
            std::list< WorkAdapt* > tl ;
            if(is_finalize_work()) wa->get_main_result(this);
            wa->preemptAndGetResultToJump(j);//preempted
            wa->getThiefList(tl);// Recuperation de la liste des voleurs de wa
            thief_list.splice(++thief_list.begin(), tl);
            _fw = wa->get_finalize();
            _finalize_signal = _fw->_finalize_signal;
            thief_list.pop_front();

            tmp = _progress;

            if(!tmp) {
             _signalJump = true;
              not_finpar = 0;
             _update_work = true;
            merge(); //merge with thief result
            jump(j);//Jump to the next work
          }

         pthread_mutex_unlock(&_mutex);

        return tmp;
      }

      
      void preemptAndGetResultToJump(NextJump*& j)
      {
          bool tmp1, tmp2;
          pthread_mutex_lock(&_mutex);
              _signalMerge = true;
              _progress = false;
             tmp1 = not_finpar;
             tmp2 = _signalJump;
             SetJump(j);// permet d'arreter le calcul en cours
          pthread_mutex_unlock(&_mutex);

          while(tmp1 && !tmp2){
            pthread_mutex_lock(&_mutex);
              tmp1 = not_finpar;
              tmp2 = _signalJump;
            pthread_mutex_unlock(&_mutex); 
           sched_yield();
          }
 
          if(_update_work) {
             pthread_mutex_lock(&_mutex); 
               SetJump(j);
             pthread_mutex_unlock(&_mutex);
           }
      }


      void SetJump(NextJump*& j)
      {  
          if(is_finalize_work()) {
           
            get_finalize_work(_fw);
           
            if(!is_empty()) {
              _finalize_signal = true;
              _fw->_finalize_signal = true;
            }

          }
          setjump(j);
      }
    
      void Jump(NextJump*& j){

        pthread_mutex_lock(&_mutex);
          jump(j); //jump
        pthread_mutex_unlock(&_mutex);
      } 
   
      FinalizeWork* get_finalize() { return _fw;}
 
      void  getThiefList( std::list< WorkAdapt* >& tl)
      {
          pthread_mutex_lock(&_mutex);
             tl = thief_list;
          pthread_mutex_unlock(&_mutex);
      }


      int nb_thief() { return thief_list.size(); }

     
      
      void set_signal_jump() {
          pthread_mutex_lock(&_mutex); 
            _signalJump = true;
          pthread_mutex_unlock(&_mutex);
      }

      void Run();

      void local_main_computation(WorkAdaptPtr& me);

     friend class AdaptSteal;
};


//Local computation
template <class WorkToCompute>
void local_computation(WorkToCompute& me) {
   while(me->ExtractNextSeq())  me->LocalCompute();
}


/*
 * Adaptive engine for object inheriting from Work
 */
void  FinalizeSteal(FinalizeWorkPtr& me) {

     FinalizeWorkPtr wp2;
        
     if(me->ExtractPar(wp2)) {
        cilk_spawn local_computation(wp2);//nanoloop computation    
        cilk_spawn FinalizeSteal(me);
        cilk_spawn FinalizeSteal(wp2);
         
        cilk_sync;
     }
};

/*
 Finalize work
*/
void Finalize(FinalizeWorkPtr& me) {

         cilk_spawn local_computation(me);//nanoloop computation
         cilk_spawn FinalizeSteal(me);
        
         cilk_sync;
};

void  AdaptSteal(WorkAdaptPtr& me);

void local_steal_computation(WorkAdaptPtr& wp2) {
  
    bool tojump=true;
    int nbjump=0;
    while(tojump){  

      //cilk_spawn local_computation(wp2);//nanoloop
      while(wp2->ExtractNextSeq())  wp2->LocalCompute();//nanoloop
      if(wp2->MustJump()){
        NextJump *j ;
          if(wp2->getResultToPreempt(j)) {
           wp2->merge();
           wp2->Jump(j);
           //nbjump++;
          } else { wp2->set_signal_jump(); tojump=false;}
        } else  { wp2->set_signal_jump(); tojump=false; }
     }//end microloop


    //Work To finalize after preemption
    if(wp2->_finalize_signal) {

         std::cout << "Preemption and finalization..." << std::endl;
         FinalizeWork *fw = wp2->get_finalize();
         FinalizeWorkPtr fw_ptr(fw);
         cilk_spawn Finalize(fw_ptr);//Local task
    }

    cilk_spawn AdaptSteal(wp2);
}

/*
 Steal work
*/
void AdaptSteal(WorkAdaptPtr& me)
{
        WorkAdaptPtr wp2;
         if(me->ExtractPar(wp2)) {
           cilk_spawn local_steal_computation(wp2);
           cilk_spawn AdaptSteal(me);
         }
};

void WorkAdapt::local_main_computation(WorkAdaptPtr& me) {
    bool go_on ;
    bool no_spawn = true;
    int nbjump=0;
    do {
        go_on = me->extract_next_macrostep() ;
        {
          bool tojump=true;
          while(tojump){
            //cilk_spawn local_computation(me); //nanoloop
            while(me->ExtractNextSeq())  me->LocalCompute();//nanoloop
            //cilk_spawn AdaptSteal(me); // task to steal
            NextJump *j ;
            if(me->MainJump()){
              if(me->getResultToPreempt(j)) {
                if(me->is_finalize_work()){
                  if(!(me->_finalize_signal)) {
                    FinalizeWork *fw = me->get_finalize();
                    FinalizeWorkPtr fw_ptr(fw);
                    cilk_spawn Finalize(fw_ptr);
                    cilk_sync;
                   }
                 }
                  me->merge();
                  me->Jump(j);
                  nbjump++;
             } else tojump=false;
            } else tojump=false;
           }//end microloop
          //if(nbjump)   std::cout << "main_nbjump=" << nbjump << std::endl;
        }

    } while( go_on );
      
      if(nbjump)   std::cout << "main_nbjump=" << nbjump << std::endl;
}


void Main_Task(WorkAdapt *current_work) {
    WorkAdaptPtr me ( current_work, unmanaged() );

    cilk_spawn current_work->local_main_computation(me);
    cilk_spawn AdaptSteal(me);

    cilk_sync;
}

inline void WorkAdapt::Run()
{
  Main_Task(this); 
}

inline void run_adapt(adapt::WorkAdapt* work)
{
  work->Run();   
};

} // en adapt

//#endif // _CILK_ADAPT_H_
