#ifndef _ATHA_Prod_iter_H_
#define _ATHA_Prod_iter_H_

#include <numeric>
#include "../../tbb_adapt.h"
#include "tbb/partitioner.h"
#include "../../util/getoptions.h"

std::size_t macroloop_initsize = 300 ; //defaul value

/*
*/
struct scan_tag {
    static bool is_final_scan() {return true;}
};

struct transform_tag {
    static bool is_final_scan() {return false;}
};


template<typename Range>
class MyNextJump : public adapt::NextJump {

     typedef typename Range::const_iterator range_type;

     range_type _beg;
     range_type _end;
   public :
     MyNextJump(range_type beg, range_type end): NextJump(), _beg(beg), _end(end) { }
     void init(range_type beg, range_type end) { _beg = beg; _end = end; }
     range_type get_end() const { return _end;}
     void set_end(range_type end){ _end = end;}
     range_type  get_beg() const { return _beg;}
     void set_beg(int beg) {  _beg = beg;}
     bool is_empty() const { return _end <= _beg; }
    // void init_work() const { cout << "MyNextjump" << endl; }
};

/*
 Finalize class
*/
template<typename Range, typename Body>
class MyFinalizeWork : public adapt::FinalizeWork {

    typedef typename Range::const_iterator range_type;
    typedef typename std::size_t size_type;


    Range _range;
    Body*  _body;
    size_type _c;
    size_type _pargrain;
    size_type _seqgrain;
    range_type _beg;
    range_type _end;
    range_type _beg_local;
    range_type _end_local;
    Range _local_range;

    public :
     MyFinalizeWork(Range& range, Body*& body, size_type c=1, size_type pargrain=2): FinalizeWork(), 
       _range(range), _body(body), _c(c), _pargrain(pargrain), _beg(range.begin()), _end(range.end()) 
     {
          if(c <=0) _seqgrain = 1;
          else _seqgrain = _c * adapt::log_2(size_type(_end-_beg)+2);
     }
 

     bool extract_nextseq()
    {

       if( _beg < _end ) {
            _beg_local = _beg; _beg = _beg + _seqgrain;
            if( _beg > _end)  _beg = _end;
            _end_local = _beg;
          _local_range = Range(_beg_local, _end_local);
        return true;
       }

       return false;

    }


    bool extract_par( FinalizeWork*& sw)
    {
      if(size_type(_end - _beg) > _pargrain) {

             Range r(_beg, _end);
             Range r_thief(r, tbb::split());
             Body*  bthief = new Body(*_body, tbb::split());
     
             _range = r;
             _end = _range.end();
             sw = new  MyFinalizeWork<Range, Body>(r_thief, bthief, _c);
        //std::cout << "transform steal work : (i, j)=" << "("<< _range.begin() << ", " << _range.end() << ")" << std::endl;
            return true;
       }
       else {
            sw = NULL;
           return false;
       }

    }

 
    void localcompute()
    {
      (*_body)(_local_range, transform_tag());
    }

};

/*
 This class adapt the PrefixWork to the semantic of the parallel_scan of tbb
 by using Range, Body, and Partitioner
*/ 
template<typename Range, typename Body>
class PrefixWork : public adapt::WorkAdapt
{
    Body* _body;
    Range _range;
    Range _local_range;
    typedef typename Range::const_iterator range_type;
    typedef typename std::size_t size_type;

    const range_type _init_beg;
    const range_type _init_end;
    size_type _seqgrain;
    size_type _pargrain;
    size_type _c;
    range_type _beg;
    range_type _end;
    range_type _beg_local;
    range_type _end_local;
    range_type _work_do_size;

    public:

     PrefixWork(const Range& range, Body* body, size_type c =1, size_type pargrain=2): 
       WorkAdapt(), _range(range), _body(body), _init_beg(range.begin()), _init_end(range.end()), _pargrain(pargrain), _c(c)
    {
            _beg = _range.begin();
           _end = _range.end();
           if(c <=0) _seqgrain = 1;
           else _seqgrain = _c * adapt::log_2(size_type(_end-_beg)+2);
    }


    bool extract_nextseq()
    {

        if( _beg < _end ) {
            _beg_local = _beg; _beg +=_seqgrain;
            if( _beg > _end)  _beg = _end;
            _end_local = _beg;
          _local_range = Range(_beg_local, _end_local);
        return true;
         } 
       return false;

    }


    bool extract_par( WorkAdapt*& sw)
    {
           if(size_type(_end - _beg) > _pargrain) {
             
             Range r(_beg, _end);
             Range r_thief(r, tbb::split());
             Body*   bthief = new Body(*_body, tbb::split());
             _range = r;
             _end = _range.end();
             sw = new  PrefixWork<Range, Body>(r_thief, bthief, _c);
             
             //std::cout << "after work : (i, j)=" << "("<< _range.begin() << ", " << _range.end() << ")" << std::endl;
             //std::cout << "stolen work : (i, j)=" << "("<< r_thief.begin() << ", " << r_thief.end() << ")" << std::endl;
            return true;
           }
           else {
            sw = NULL;
           return false;
          }
    }

     bool  extract_next_macrostep()
     {
          static std::size_t alpha = macroloop_initsize ;
          bool extractmacro = size_type(_beg  < _init_end);
          if(extractmacro){

            std::size_t completed_size = size_type(_beg - _init_beg) ;
            std::size_t lg_sz = adapt::log_2(adapt::log_2( completed_size + 4 ) ) ;
            std::size_t size_macrostep= alpha*completed_size / lg_sz ;
            if (size_macrostep < macroloop_initsize)
                size_macrostep = macroloop_initsize ;
            _end = _beg + size_macrostep ;

            if(_end > _init_end) _end = _init_end;
            
            Range r(_beg, _end);
            _range = r;
             std::cout << "Macrostep:  _beg = " << (_beg-_init_beg) << ", _end=" << (_end-_init_beg) << std::endl;
          }
        return extractmacro;
      }


    range_type get_end() const
    {
      return _end;
    }


    void join(const WorkAdapt* jw)
    {
        const  PrefixWork * src = dynamic_cast< const  PrefixWork*> ( jw ); 
        
         _body->join(*(src->_body));    
        
          range_type tmp = src->get_end();
          std::cout << "2 tmp = " << tmp << std::endl;
         _body->assign(--tmp);
    }

   void  get_main_result(const WorkAdapt* jw ) 
   {
        const  PrefixWork * src = dynamic_cast< const  PrefixWork*> ( jw );
        _body->get_main_result(*(src->_body));
   }

    void  localcompute()
    { 
      (*_body)(_local_range, scan_tag());
    }

    inline bool is_finalize_work() const { return true; }

    void get_finalize_work(adapt::FinalizeWork*& fw)
    {
        range_type tmp = _beg;
        Range r(_init_beg, --tmp);
        std::cout << "1 tmp = " << tmp << std::endl;
        fw = new MyFinalizeWork<Range, Body>(r, _body, _c);
    }


    void setjump(adapt::NextJump*& j) {

          range_type end = _end;
          _end = _beg;

          j = new MyNextJump<Range>(_beg, end);    
   }

   void jump(const adapt::NextJump* j) {
        
        const  MyNextJump<Range>* nextj = dynamic_cast< const  MyNextJump<Range>*> ( j );
        _beg = nextj->get_beg();
        _end = nextj->get_end();
       //std::cout << "Jump work =(" << _beg << "," << _end << ")" << std::endl;
   }

    Util::OStream& pack(Util::OStream& os) const
    {
        return os ;
    }

    Util::IStream& unpack(Util::IStream& is, WorkAdapt*& target)
    {
        return is;
    }

};

template<typename Range, typename Body>
void parallel_scan_adapt(const Range& range, Body& body, size_t seqgrain=100, size_t pargrain=2) 
{
     PrefixWork<Range,Body>* work = new  PrefixWork<Range,Body>(range, &body, seqgrain);
     adapt::run_adapt(work);
}

#endif
