#ifndef _TBB_PREFIX_H_
#define _TBB_PREFIX_H_

#include <numeric>
#include "../cilk_adapt.h"

unsigned int macroloop_initsize = 300 ; //defaul value

class MyNextJump : public adapt::NextJump {

     int _beg;
     int _end;
   public :
     MyNextJump(int beg=0, int end=0): NextJump(), _beg(beg), _end(end) { }
     void init(int beg, int end) { _beg = beg; _end = end; }
     int get_end() const { return _end;}
     void set_end(int end){ _end = end;}
     int  get_beg() const { return _beg;}
     void set_beg(int beg) {  _beg = beg;}
     bool is_empty() const { return _end <= _beg; }
    // void init_work() const { cout << "MyNextjump" << endl; }
};

template <class T>
struct my_fct {
    T _val;
    my_fct(T val):_val(val) { };
    T operator()(T a){
         return  (_val + a);
    }
};


template<class T>
class MyFinalizeWork : public adapt::FinalizeWork {
    T* _first;
    T* _last;
    int _beg;
    int _end;
    T _val;
    int _beg_local ;
    int _end_local;
    int _c;
    int _pargrain;
    int _seqgrain;

    public :
     MyFinalizeWork(T* first, T* last, int beg, T val=0, int c=1, int pargrain=2): FinalizeWork(), 
       _first(first),  _last(last), _beg(beg), _end(last-first), _val(val), _c(c), _pargrain(pargrain) 
     { 
        if(c <=0) _seqgrain = 1;
        else _seqgrain = _c * adapt::log_2(_end);
       /* if(val){ cout << "Finalize intervall :[" << _beg << " " << _end << "[" << endl;
                  cout << "My Main val = " << _val << endl;
        }*/
     }
 

     bool extract_nextseq()
    {

        if( _beg < _end ) {
            _beg_local = _beg; _beg +=_seqgrain;
            if( _beg > _end)  _beg = _end;
            _end_local = _beg;
           return true;
         }
       return false;

    }


    bool extract_par( FinalizeWork*& sw)
    {
           if((_end - _beg) > _pargrain) {
             int i, j;
             int mid = _end - ((_end - _beg)/2);


             i = mid;
             j = _end;
             _end = mid;
            //cout << "work : (i, j)=" << "("<< _beg << ", " << _end << ")" << endl;
            //cout << "stolen work : (i, j)=" << "("<< i << ", " << j << ")" << endl;
             sw = new  MyFinalizeWork<T>( _first, _first+j, i, _val, _c);
            return true;
           }
           else {
            sw = new  MyFinalizeWork<T>(_first, _first, 0, 0, 0);
           return false;
          }
     }

 
     void localcompute()
     {
       //cout << "Finalize intervall in Localcompute:[" << _beg_local << " " << _end_local << "[" << endl;
       //cout << "My Main val = " << _val << endl;
       //Util::Thread::sleep(1, 1000);
       std::transform(_first + _beg_local, _first + _end_local, _first + _beg_local, my_fct<T>(_val)) ;
     }

};

template<class T>
T partial_sum_local_compute (T first, T last, T res) {
       typedef typename std::iterator_traits<T>::value_type val_t;
       val_t val =  *res ;
       while(++first!=last){
           val=(val+*first);
           *++res=val;
       }
  return res;
};


 
template<class T>
class PrefixWork : public adapt::WorkAdapt
{
     T* _first;
     T* _last;
     T* _res;
     T _local_res;
    int _beg;
    int _end;
    const int _init_beg;
    const int _init_end;
    int _seqgrain;
    int _pargrain;
    int _c;
    int _beg_local;
    int _end_local;
    int _work_do_size;
    T _main_res;
    T _thieft_res;
    public:

     PrefixWork(T* first, T* last, T* res, int beg, int c =1, int pargrain=2): 
       WorkAdapt(), _first(first),  _last(last), _res(res), _local_res(T(0)),
         _beg(beg), _init_beg(beg), _init_end(last - first), _pargrain(pargrain), _c(c)
    {
            _end = _init_end;
           if(c <=0) _seqgrain = 1;
           else _seqgrain = _c * adapt::log_2(_end);
    }


    bool extract_nextseq()
    {
        
        if( _beg < _end ) {
            _beg_local = _beg; _beg +=_seqgrain;
            if( _beg > _end)  _beg = _end;
            _end_local = _beg;
           return true;
         } 
       return false;

    }


    bool extract_par( WorkAdapt*& sw)
    {
           if((_end - _beg) > _pargrain) {
             int i, j;
             int mid = _end - ((_end - _beg)/2);

             
             i = mid;
             j = _end;
             _end = mid;
             //cout << "work : (i, j)=" << "("<< _beg << ", " << _end << ")" << endl;
             //cout << "stolen work : (i, j)=" << "("<< i << ", " << j << ")" << endl;
             sw = new  PrefixWork<T>(_first, _first+j, _res, i, _c);
            return true;
           }
           else {
            sw = new  PrefixWork<T>(_first, _first, _res, 0, 0);
           return false;
          }
    }

     bool  extract_next_macrostep()
     {
          static unsigned int alpha = macroloop_initsize ;
          bool extractmacro = (_beg  < _init_end);
          if(extractmacro){

            unsigned int completed_size = _beg - _init_beg ;
            int lg_sz = adapt::log_2(adapt::log_2( completed_size + 4 ) ) ;
            int size_macrostep= alpha*completed_size / lg_sz ;
            if (size_macrostep < macroloop_initsize)
                size_macrostep = macroloop_initsize ;
            _end = _beg + size_macrostep ;

            if(_end > _init_end) _end = _init_end;
             //std::cout << "Macrostep:  _beg = " << _beg << ", _end=" << _end << std::endl;
          }
        return extractmacro;
      }

    bool is_empty() {
          return (_beg >= _end);
    }

    void join(const WorkAdapt* jw)
    {
        const  PrefixWork * src = dynamic_cast< const  PrefixWork*> ( jw ); 
    
          T thief_res = src->get_result();
           //std::cout << "victim_res =" << _local_res << std::endl;
           //std::cout << "thief_res =" << thief_res << std::endl;
          _local_res +=thief_res ;
          int thieft_end = src->get_end();
           *(_res+(thieft_end-1))   = _local_res;
    }

   void  get_main_result(const WorkAdapt* jw ) 
   {
        const  PrefixWork * src = dynamic_cast< const  PrefixWork*> ( jw );
        _main_res = src->get_result();
        std::cout << "main_res =" << _main_res << std::endl;
   }

    void  localcompute()
    { 
       //Util::Thread::sleep(1, 1000);
       T val = ( (_local_res==T(0))? *(_first+_beg_local) : (_local_res + *(_first+_beg_local)));
      *(_res+_beg_local)   = val ; 
       _local_res = *(partial_sum_local_compute(_first + _beg_local, _first + _end_local, _res+_beg_local)) ;
    }

    inline bool is_finalize_work() const { return true; }

    void get_finalize_work(adapt::FinalizeWork*& fw)
    { 
        fw = new MyFinalizeWork<T>(_res, _res+(_beg-1), _init_beg, _main_res, _c);
    }

    T get_result() const
    {
        return _local_res ;
    }

    int get_end() const 
    {
      return _end;
    }

    void setjump(adapt::NextJump*& j) {
        
          int end = _end;
          _end = _beg;

          j = new MyNextJump(_beg, end);   
                
   }

   void jump(const adapt::NextJump* j) {
        
        const  MyNextJump * nextj = dynamic_cast< const  MyNextJump*> ( j );
        _beg = nextj->get_beg();
        _end = nextj->get_end();
       /*if(_beg < _end) {
          cout << "Jump Intervall to complete :[" << _beg << " " << _end << "[" << endl;
       }*/
   }

};

template <class T>
T adapt_prefix(T* first, T* last, T* res,
      int seqgrain=10, int pargrain=2) 
{
    /* init remote iterator */
     PrefixWork<T>* work = new  PrefixWork<T>(first, last, res, 0, seqgrain);
    adapt::run_adapt(work);
    T val_end = work->get_result();
    return val_end ; 
}

#endif
