
#include <iostream>
#include <numa.h>
#include <numaif.h>
#include <sys/mman.h>
#include <signal.h>

#include "tbb/task_scheduler_init.h"
#include "tbb/tick_count.h"
#include "../merge/merge.h"

typedef double val_t ;

inline double gettime()
{
   return Util::WallTimer::gettime();
}

template<class T>
bool is_sorted(T* output, int sz)
{
     bool is_sorted = true;
     for(int i=1; i < sz; i++) {
         if(output[i-1] > output[i]) {
             std::cout << "output[" << i-1 << "]=" << output[i-1] << " is not less than  ";
             std::cout << "output[" << i << "]=" << output[i] << ", " << std::endl;
             std::cout << "Then the table is not sorted well" << std::endl;
             is_sorted = false;
           break;
        }
     }
    //if(is_sorted) std::cout << a1::System::getRank() << ":The table has been well sorted..........";
    //std::cout << std::endl;
return is_sorted;
}



int seuil = 10000;

template<class T>
class MERGE_ADAPT_TASK : public tbb::task{

   T* A1;
   T* A2;
   T* B1;
   T* B2;
   T* input;
   int c;

  public:

   MERGE_ADAPT_TASK(T* _A1, T* _A2, T* _B1, T* _B2,  T* _input, int _c): A1(_A1), A2(_A2), 
                         B1(_B1), B2(_B2), input(_input), c(_c){}

   tbb::task* execute() {
     adapt_merge(A1 , A2, B1, B2, input, c);
    return NULL;
   }
};

template<class T>
class PMERGE_SORT : public tbb::task{
 
   T* input;
   T* tmp;
   size_t sz;
   int c;
   
  public:

   PMERGE_SORT(T* _input, T* _tmp, size_t _sz, int _c): input(_input), tmp(_tmp), sz(_sz), c(_c){}


  tbb::task* execute() {
    if(sz <= seuil) stable_sort(input, input+sz, less_than<T>());
    else {
          int quarter = (sz/4);
          T* A, *B, *C, *D, *tmpA, *tmpB, *tmpC, *tmpD;

          A = input;
          tmpA = tmp;
          B = A + quarter;
          tmpB = tmpA + quarter;
          C = B + quarter;
          tmpC = tmpB + quarter;
          D = C + quarter;
          tmpD = tmpC + quarter;

          PMERGE_SORT<T>& thief1 = *new(allocate_child()) PMERGE_SORT<T>(A, tmpA, quarter, c);
          PMERGE_SORT<T>& thief2 = *new(allocate_child()) PMERGE_SORT<T>(B, tmpB, quarter, c);   
          PMERGE_SORT<T>& thief3 = *new(allocate_child()) PMERGE_SORT<T>(C, tmpC, quarter, c);
          PMERGE_SORT<T>& thief4 = *new(allocate_child()) PMERGE_SORT<T>(D, tmpD, sz-3*quarter, c);
          set_ref_count(5);
          spawn(thief1);
          spawn(thief2);
          spawn(thief3); 
          spawn_and_wait_for_all(thief4);
 
          MERGE_ADAPT_TASK<T>& thief5 = *new(allocate_child()) MERGE_ADAPT_TASK<T>(A, A+quarter, B, B+quarter, tmpA, c);
          MERGE_ADAPT_TASK<T>& thief6 = *new(allocate_child()) MERGE_ADAPT_TASK<T>(C, C+quarter, D, input+sz, tmpC, c);
          set_ref_count(3);
          spawn(thief5);
          spawn_and_wait_for_all(thief6);

          MERGE_ADAPT_TASK<T>& thief7 = *new(allocate_child()) MERGE_ADAPT_TASK<T>(tmpA, tmpC, tmpC, tmpA+sz, A, c);
          set_ref_count(2);
          spawn_and_wait_for_all(thief7);
   }
  
  return NULL;
 }
};

template<class T>
inline void tbb_adaptive_sort(T* input, T* tmp, size_t size, int c)
{
  PMERGE_SORT<T>& sort_task = *new(tbb::task::allocate_root()) PMERGE_SORT<T>(input, tmp, size, c);
  tbb::task::spawn_root_and_wait(sort_task);
}


int main(int argc, char ** argv) {
      

      int size = (argc > 1)?atoi(argv[1]):100000000;
      int nthread = (argc > 2)?atoi(argv[2]):1;
      int c = (argc > 3)?atoi(argv[3]):100;
      int seuil = (argc > 4)?atoi(argv[4]):10000;
      int ch = (argc > 5)?atoi(argv[5]):1;


      unsigned long numa_node = 255;
      val_t *data = (val_t *)mmap(0, (size)*sizeof(val_t),PROT_READ|PROT_WRITE,
                                 MAP_PRIVATE|MAP_ANONYMOUS,0,0);
      if (mbind(data,size*sizeof(val_t),MPOL_INTERLEAVE,&numa_node,64,0) != 0)
      {
          perror("mbind");
      }
      
      val_t *tmp = (val_t *)mmap(0, (size)*sizeof(val_t),PROT_READ|PROT_WRITE,
                                 MAP_PRIVATE|MAP_ANONYMOUS,0,0);
      if (mbind(tmp,size*sizeof(val_t),MPOL_INTERLEAVE,&numa_node,64,0) != 0)
      {
          perror("mbind");
      }

      std::cout << "********start computation with size =" << size << "*********" << std::endl;
       
       srand48(0);
       for(int i =0; i < size; i++) data[i] = drand48();

       tbb::task_scheduler_init init(nthread);

       tbb::tick_count t0, t1;
       t0 = tbb::tick_count::now();
       if(ch==1) {
          std::cout << "stable sort adaptatif" << std::endl;
          tbb_adaptive_sort(data, tmp, size, c); 
       } else {
          std::cout << "stable sort sequentiel STL" << std::endl;
          std::stable_sort(data, data+size); 
       }
       t1 = tbb::tick_count::now();

       std::cout << "-----------------------------------" << std::endl;
       printf (": time = %.6f s\n", (t1-t0).seconds());
       std::cout << std::endl;
       std::cout << "-----------------------------------" << std::endl;

       if(is_sorted(data, size)) 
         std::cout << a1::System::getRank() << ":The table has been well sorted.........." << std::endl;

       numa_free(data, size*sizeof(val_t));  
       numa_free(tmp, size*sizeof(val_t)); 
  return 0;
};

