/*Author : D Traore
 *
*/

#ifndef _TRANSFORM_H_
#define _TRANSFORM_H_

#include <iostream>
#include <assert.h>
#include <algorithm>

#include <cilk.h>

typedef int IndexType;

template <class In, class Out, class op> class WorkAdapt ;


inline uint log_2 ( uint a )  // returns [log_2 a]+1
{
        uint res = 0 ;
        for ( ; a != 0 ; a >>=1, res++ ) ;
        return res ;
}


template<class In, class Out, class Op >
class WorkAdapt {

    In _first;
    In _last;
    Out _first_out;
    Op _op;
    IndexType _beg;
    IndexType _end;
    IndexType _seqgrain;
    IndexType _pargrain;
    IndexType _c;
    IndexType _beg_local;
    IndexType _end_local;
    
    pthread_mutex_t  _mutex ;

    public :
    WorkAdapt()
        : _first(), _last(), _first_out(), _op(),
          _beg(0), _end(0), _seqgrain(0), _pargrain(0), _c(0),
          _beg_local(0), _end_local(0), _mutex()
    { 
             pthread_mutex_init(&_mutex, NULL);
    }

    WorkAdapt(In first, In last, Out first_out, Op op,
              IndexType beg, IndexType c =1, IndexType pargrain=2)
    : _first(first),  _last(last), _first_out(first_out), _op(op),
      _beg(beg), _pargrain(pargrain), _c(c)
    {
        _end = last - first;
         if(c <= 0) /*_c = c = 1*/ _seqgrain = -c + 1;
         else _seqgrain = c  * log_2(_end);
         pthread_mutex_init(&_mutex, NULL);
    }

    ~WorkAdapt()
    {
    }

    bool is_empty()
    {
        return _end <= _beg;
    }
    void verif() const {
        std::cout << "verif: ( " << _beg << ", " << _end << " )" << std::endl;
    }

    bool extract_nextseq()
    {
        bool local_extract_is_possible = false;
       pthread_mutex_lock(&_mutex);
        if( _beg < _end ) {
            _beg_local = _beg;
            _beg += _seqgrain;
            if( _beg > _end)  _beg = _end;
            _end_local = _beg;
            local_extract_is_possible = true;
        }
       pthread_mutex_unlock(&_mutex);
      return local_extract_is_possible;
    }

    bool extract_par(WorkAdapt<In,Out,Op>*& sw)
    {
        bool steal_extract_is_possible = false;
       pthread_mutex_lock(&_mutex);
        if((_end - _beg) > _pargrain) {
            IndexType i, j;
            IndexType mid = _end - ((_end - _beg)/2);
            i = mid;
            j = _end;
            _end = mid;

            steal_extract_is_possible = true;
#ifdef PRINT
            std::cout << "work : (i, j)=" << "("<< _beg << ", " << _end << ")" << std::endl;
            std::cout << "stolen work : (i, j)=" << "("<< i << ", " << j << ")" << std::endl;
#endif
            sw = new WorkAdapt<In, Out, Op>(_first, _first+j, _first_out, _op, i, _c);
        }
        else {
            sw = 0;
        }
      pthread_mutex_unlock(&_mutex);   
  return steal_extract_is_possible;

    }

    void  localcompute()
    {
#ifdef PRINT
        Util::logfile() << "starting local compute" << std::endl;
#endif
        std::transform(_first + _beg_local ,_first + _end_local, _first_out + _beg_local , _op );

#ifdef PRINT
        Util::logfile() << "stoping local compute" << std::endl;
#endif
    }

};

//Local computation 
template <class InputIterator, class OutputIterator, class UnaryOperator>
void local_computation(WorkAdapt<InputIterator, OutputIterator, UnaryOperator> *wa) {
   while(wa->extract_nextseq())  wa->localcompute();
}

//Steal work
template <class InputIterator, class OutputIterator, class UnaryOperator>
void stealwork(WorkAdapt<InputIterator, OutputIterator, UnaryOperator> *vw)
{ 
    
    WorkAdapt<InputIterator, OutputIterator, UnaryOperator> *sw = 0; 
     
    if(vw->extract_par(sw)) {
    
         cilk_spawn local_computation(sw);//nanoloop computation
    
         cilk_spawn stealwork(vw);//stealable task
         cilk_spawn stealwork(sw); //stealable task
         
         cilk_sync;
    } 
        
};

/////////////////////////////////////////////////////////////////////////////////////////

template <class InputIterator, class OutputIterator, class UnaryOperator>
void my_unary_transform(InputIterator begin, InputIterator end, OutputIterator to_fill,
                     UnaryOperator op,  size_t seqgrain, size_t pargrain ) {


     WorkAdapt<InputIterator, OutputIterator, UnaryOperator> *wa = new WorkAdapt<InputIterator, OutputIterator, UnaryOperator>(begin, end, to_fill, op, 0, seqgrain);


     cilk_spawn local_computation(wa);
     cilk_spawn stealwork(wa);

     cilk_sync;
};

template <class InputIterator, class OutputIterator, class UnaryOperator>
void transform(InputIterator begin, InputIterator end, OutputIterator to_fill , UnaryOperator op,  size_t seqgrain=100, size_t pargrain=2 )
{
 my_unary_transform<InputIterator, OutputIterator, UnaryOperator>(begin, end, to_fill, op, seqgrain,  pargrain);
}

#endif // _TRANSFORM_H_
